#!/usr/bin/perl
#
# I carved this out so that I could start multiple times in parallel.
# I do not have to close the filehandle that refers to each instance
# of this script, but this accumulates defunct processes that clear
# only when the main script completes. To avoid the main script hanging
# I must make sure that a flag is set every way this script can complete
# or fail to complete.  I do this by touching a file called "done.flag".

#
# Pragmata
# ````````
use 5.006;
use strict;
use warnings;


# Load Perl modules - standard modules only
# `````````````````````````````````````````
use POSIX;
use Storable;
use File::Spec;
use File::Copy;
use Data::Dumper;
use File::Basename;


# Global variables are all UPPERCASE throughout
# `````````````````````````````````````````````
our $MAIN = basename($0);
our $DONE = "touch done.flag";
#our $HOME = File::Spec -> catfile($ENV{CSPP_NUCAPS_HOME}, $ENV{CSPP_NUCAPS_REVNAME}, 'CrISOPS');
our $HOME = File::Spec -> catfile($ENV{CSPP_NUCAPS_HOME}, $ENV{CSPP_NUCAPS_REVNAME}, 'OPS');
(-d $HOME) or do { system($DONE);
  print "$MAIN is designed only to be called from within run_nucaps.pl !\n";
  die "$MAIN\:: CSPP_NUCAPS subdirectory $HOME not found.\nAbort"; };


# Recover hashes stored by process caller
# ```````````````````````````````````````
my $pwd = getcwd();
(-e '../options.store') or do { system($DONE);
  print "$MAIN is designed only to be called from within run_nucaps.pl !\n";
  die "$MAIN\:: cannot find ../options.store\nAbort"; };
my %options = %{retrieve('../options.store')};
(-e 'granule.store') or do { system($DONE);
  print "$MAIN is designed only to be called from within run_nucaps.pl !\n";
  die "$MAIN\:: cannot find granule.store\nAbort"; };
my %granule = %{retrieve('granule.store')};
my $wrkdir = basename(dirname($pwd));
my $instance = basename($pwd);
#pretty_dump('%options', %options);
#pretty_dump('%granule', %granule);


# Loop over the multi-step process
# ````````````````````````````````
foreach my $step ( qw/preproc retrieval subset/ ) {
  my $skipretrieval = 0; # indicates skip retrieval, but only during retrieval step
  (($step eq 'retrieval') and $options{skipretrieval}) and $skipretrieval = 1;
  (($step eq 'subset') and !$options{subset}) and last;
  # Create the perl script name for this step
  my $proc = sprintf("NUCAPS_%s.pl", ucfirst($step));
  # Make links to the support files
  foreach (sort(glob("../../support/$step/*"))) {
    my $linkname = basename($_);
    (-e $linkname) and next;
    symlink($_, $linkname) or do { system($DONE);
      die "$MAIN\:: Cannot make link to $_\nAbort"; };
  }
  # Make change to filenames we cannot know until after NUCAPS_Preproc.pl
  if ($step eq 'retrieval') {
    foreach ( qw/BOUNDING_BOX_PRODUCT NUCAPS_ALL_PRODUCT/ ) {
      # Should only be one of each, but just in case we take the last
      $granule{$_} = (sort(glob($granule{$_})))[-1];
    }
  }
  #
  # Abort on error because this really means the system is broken
  # `````````````````````````````````````````````````````````````
  open(my $fhI, "<../../$proc.PCF.template") or do { system($DONE);
    die "$MAIN\:: Cannot open $proc.PCF.template for read.\nAbort"; };
  open(my $fhO, ">$proc.PCF") or do { system($DONE);
    die "$MAIN\:: Cannot open $wrkdir/$instance/$proc.PCF for write.\nAbort"; };
  while (<$fhI>) { chomp;
    foreach my $key (sort keys %granule) {
      #
      # $key in %granule marks content to replace in PCF template
      # `````````````````````````````````````````````````````````
      $_ =~ s|<$key>|$granule{$key}|g;
      #
      # If $granule{$key} is a file in the parent directory, make a link
      # ````````````````````````````````````````````````````````````````
      my $linkname = $granule{$key};
      (-e $linkname) and next;
      my $target = File::Spec -> catfile('..', $linkname);
      (-e $target) or next;
      symlink($target, $linkname) or do { system($DONE);
        die "$MAIN\:: Cannot make link to $target\nAbort"; };
    }
    print $fhO "$_\n";
  }
  close($fhI) or do { system($DONE);
    die "$MAIN\:: Cannot close $proc.PCF.template.\nAbort"; };
  close($fhO) or do { system($DONE);
    die "$MAIN\:: Cannot close $wrkdir/$instance/$proc.PCF.\nAbort"; };
  #
  # Move PCF files into audit trail versions with unique filenames
  # ``````````````````````````````````````````````````````````````
  # Maybe the thing to do here is to name like the output .nc files with date/time?
  my $cmd = "perl $HOME/scripts/$proc";
  my $jco = "grep job_coverage $proc.PCF >> $proc.log";
  my $avn = "grep AVN_FILE $proc.PCF >> $proc.log";
  open(my $fh, ">$proc.bash") or do { system($DONE);
    die "$MAIN\:: Cannot open file $proc.bash $!"; };
  print $fh "#!/bin/bash\n#NUCAPS execution line run by $MAIN\nulimit -s unlimited\n$cmd\n";
  close($fh) or do { system($DONE);
    die "$MAIN\:: Cannot close file $proc.bash $!"; };
  system("chmod 755 $proc.bash");
  if ($options{nothing} or $skipretrieval) {
    ($options{verbose} >= 2) and print "Skipping ... $proc for $wrkdir $instance\n";
  } else {
    ($options{verbose} >= 2) and print "Running ... $proc for $wrkdir $instance\n";
    system($cmd) == 0 or do {
      ($step eq 'subset') and last; # You do not need this warning for NUCAPS_Subset.pl
      ($options{verbose} >= 1) and do {
        warn "WARNING: $proc failed for $pwd";
        print "INFO: This fail can occur if trying to process missing FOVs\n";
        last; # if NUCAPS_Preproc.pl fails, do not run NUCAPS_Retrieval.pl
      };
    };
    system($jco);
    system($avn);
  }
}
my @copyfiles = ();
push(@copyfiles, sort(glob("*.log")));
push(@copyfiles, sort(glob("*.PCF")));
foreach (@copyfiles) {
  my @parts = split(/\./, $_);
  my $ext = pop(@parts);
  push(@parts, sprintf("%s_%s", $wrkdir, $instance));
  my $target = File::Spec -> catfile('..', join('.', @parts, $ext));
  copy($_, $target) or do {
    ($options{verbose} >= 1) and warn "WARNING: Copy failed: $!"; };
}
@copyfiles = ();
push(@copyfiles, sort(glob("NUCAPS*.nc")));
push(@copyfiles, sort(glob("NUCAPS*.txt")));
push(@copyfiles, sort(glob("*.out")));
foreach (@copyfiles) {
  copy($_, '..') or do {
    ($options{verbose} >= 1) and warn "WARNING: Copy failed: $!"; };
}
system($DONE);
exit(0);

# ---------------
  sub pretty_dump
# ===============
# pretty_dump($title, %hash);
  {
    # Do a "pretty print" of %hash by effectively grepping Dumper().
    # Assumes standard module Data::Dumper has been loaded.
    my $title = 'DEBUG '.shift;
    my (%hash) = @_;
    print '-' x length($title)."\n";
    print "$title\n";
    print '=' x length($title)."\n";
    foreach (split('\n', Dumper(\%hash))) {
      chomp;
      ((/VAR1/) or (/\}\;/)) and next;
      $_ =~ s/  //g; # shorten up the whitespace
      print "$_\n";
    }
    print "$title _END_\n";
  }
